@@ -120,7 +120,10 @@ static const char * const cmd_df_usage[] = {
 
 static char *group_type_str(u64 flag)
 {
-	switch (flag & BTRFS_BLOCK_GROUP_TYPE_MASK) {
+	u64 mask = BTRFS_BLOCK_GROUP_TYPE_MASK |
+		BTRFS_SPACE_INFO_GLOBAL_RSV;
+
+	switch (flag & mask) {
 	case BTRFS_BLOCK_GROUP_DATA:
 		return "Data";
 	case BTRFS_BLOCK_GROUP_SYSTEM:
@@ -129,6 +132,8 @@ static char *group_type_str(u64 flag)
 		return "Metadata";
 	case BTRFS_BLOCK_GROUP_DATA|BTRFS_BLOCK_GROUP_METADATA:
 		return "Data+Metadata";
+	case BTRFS_SPACE_INFO_GLOBAL_RSV:
+		return "GlobalReserve";
 	default:
 		return "unknown";
 	}
@@ -238,7 +243,7 @@ static int cmd_df(int argc, char **argv)
 	}
 	ret = get_df(fd, &sargs);
 
-	if (!ret && sargs) {
+	if (ret == 0) {
 		print_df(sargs);
 		free(sargs);
 	} else {
@@ -444,6 +449,7 @@ static int check_arg_type(char *input)
 static int btrfs_scan_kernel(void *search)
 {
 	int ret = 0, fd;
+	int found = 0;
 	FILE *f;
 	struct mntent *mnt;
 	struct btrfs_ioctl_fs_info_args fs_info_arg;
@@ -466,7 +472,6 @@ static int btrfs_scan_kernel(void *search)
 
 		if (get_label_mounted(mnt->mnt_dir, label)) {
 			kfree(dev_info_arg);
-			ret = 1;
 			goto out;
 		}
 		if (search && !match_search_item_kernel(fs_info_arg.fsid,
@@ -481,19 +486,16 @@ static int btrfs_scan_kernel(void *search)
 					space_info_arg, label, mnt->mnt_dir);
 			kfree(space_info_arg);
 			memset(label, 0, sizeof(label));
+			found = 1;
 		}
 		if (fd != -1)
 			close(fd);
 		kfree(dev_info_arg);
-		if (search)
-			ret = 0;
 	}
-	if (search)
-		ret = 1;
 
 out:
 	endmntent(f);
-	return ret;
+	return !found;
 }
 
 static int dev_to_fsid(char *dev, __u8 *fsid)
@@ -516,7 +518,7 @@ static int dev_to_fsid(char *dev, __u8 *fsid)
 
 	disk_super = (struct btrfs_super_block *)buf;
 	ret = btrfs_read_dev_super(fd, disk_super,
-			BTRFS_SUPER_INFO_OFFSET);
+				   BTRFS_SUPER_INFO_OFFSET, 0);
 	if (ret)
 		goto out;
 
@@ -550,7 +552,7 @@ static int cmd_show(int argc, char **argv)
 	char mp[BTRFS_PATH_NAME_MAX + 1];
 	char path[PATH_MAX];
 	__u8 fsid[BTRFS_FSID_SIZE];
-	char uuid_buf[37];
+	char uuid_buf[BTRFS_UUID_UNPARSED_SIZE];
 	int found = 0;
 
 	while (1) {
@@ -659,7 +661,6 @@ devs_only:
 	while (!list_empty(all_uuids)) {
 		fs_devices = list_entry(all_uuids->next,
 					struct btrfs_fs_devices, list);
-		list_del(&fs_devices->list);
 		btrfs_close_devices(fs_devices);
 	}
 out:
@@ -927,10 +928,11 @@ static int cmd_defrag(int argc, char **argv)
 }
 
 static const char * const cmd_resize_usage[] = {
-	"btrfs filesystem resize [devid:][+/-]<newsize>[gkm]|[devid:]max <path>",
+	"btrfs filesystem resize [devid:][+/-]<newsize>[kKmMgGtTpPeE]|[devid:]max <path>",
 	"Resize a filesystem",
 	"If 'max' is passed, the filesystem will occupy all available space",
 	"on the device 'devid'.",
+	"[kK] means KiB, which denotes 1KiB = 1024B, 1MiB = 1024KiB, etc.",
 	NULL
 };
 
